from sklearn.inspection import permutation_importance
import pandas as pd
import matplotlib.pyplot as plt
from boruta import BorutaPy


def impurity_decrease_importance(model, col_names):
    importances = pd.Series(model.feature_importances_, index=col_names)
    fig, ax = plt.subplots()
    fig.set_size_inches(7, 5)
    importances.plot.bar(ax=ax).get_figure()
    ax.set_title("Feature importances using MDI")
    ax.set_ylabel("Mean decrease in impurity")
    fig.tight_layout()
    return model.feature_importances_, fig


def permutations_importance(model, val_x, val_y, n_repeats):
    r = permutation_importance(model, val_x, val_y,
                               n_repeats=30,
                               random_state=1001)
    fig, ax = plt.subplots()
    pd.Series(r.importances_mean, index=val_x.columns).plot.bar(yerr=r.importances_std, ax=ax)
    ax.set_title("Feature importances using Permutations")
    fig.tight_layout()
    return r.importances_mean, fig


def boruta_selection(model, X, y, max_iter=50, verbose=0):
    boruta_selector = BorutaPy(model, n_estimators='auto', verbose=verbose, random_state=1001, max_iter=max_iter)
    boruta_selector.fit(X.values, y.ravel())
    accept = X.columns[boruta_selector.support_].to_list()
    irresolution = X.columns[boruta_selector.support_weak_].to_list()
    print(f'Accepted features: {accept}')
    print(f'Accepted features (weak): {irresolution}')
    return accept, irresolution


from hyperopt import STATUS_OK, Trials, fmin, hp, tpe
from xgboost import XGBRegressor
from functools import partial
from sklearn.metrics import mean_squared_error
from sklearn.ensemble import RandomForestRegressor

DEFAULT_XGBR_SPACE = {
    'n_estimators': hp.quniform('n_estimators', 10, 1000, 1),
    'max_depth': hp.quniform('max_depth', 3, 18, 1),
    'grow_policy': hp.choice('grow_policy', [0, 1]),
    'learning_rate': hp.quniform('learning_rate', 0.025, 0.5, 0.025),
    'booster': 'gbtree',
    'tree_method': hp.choice('tree_method', ['exact', 'approx', 'hist']),
    'gamma': hp.quniform('gamma', 0.5, 1, 0.05),
    'min_child_weight': hp.quniform('min_child_weight', 1, 6, 1),
    'subsample': hp.quniform('subsample', 0.5, 1, 0.05),
    'colsample_bytree': hp.quniform('colsample_bytree', 0.5, 1, 0.05),
    'colsample_bylevel': hp.quniform('colsample_bylevel', 0.5, 1, 0.05),
    'colsample_bynode': hp.quniform('colsample_bynode', 0.5, 1, 0.05),
}

DEFAULT_RFR_SPACE = {
        'n_estimators': hp.quniform('n_estimators', 10, 1000, 1),
        'max_depth': hp.quniform('max_depth', 3, 20, 1),
        'min_samples_split': hp.quniform('min_samples_split', 2, 100, 1),
        'min_samples_leaf': hp.quniform('min_samples_leaf', 1, 20, 1),
        'max_features': hp.quniform('max_features', 0.1, 0.9, 0.1),
        'max_samples': hp.quniform('max_samples', 0.1, 0.9, 0.1)
    }


def xgbr_score(params, **data):
    model = XGBRegressor(n_estimators=int(params['n_estimators']),
                         max_depth=int(params['max_depth']),
                         learning_rate=params['learning_rate'],
                         booster=params['booster'],
                         tree_method=params['tree_method'],
                         gamma=params['gamma'],
                         min_child_weight=int(params['min_child_weight']),
                         subsample=params['subsample'],
                         colsample_bytree=params['colsample_bytree'],
                         colsample_bylevel=params['colsample_bylevel'],
                         colsample_bynode=params['colsample_bynode'],
                         random_state=1001
                         )

    model.fit(data['t_x'], data['t_y'])
    pred = model.predict(data['v_x'])
    mse = mean_squared_error(data['v_y'], pred, squared=False)
    return {'loss': mse, 'status': STATUS_OK, 'model': model}


def rfr_score(params, **data):
    model = RandomForestRegressor(n_jobs=-1, random_state=1001,
                                  n_estimators=int(params['n_estimators']),
                                  max_depth=int(params['max_depth']),
                                  min_samples_split=int(params['min_samples_split']),
                                  min_samples_leaf=int(params['min_samples_leaf']),
                                  max_features=params['max_features'] if params['max_features'] < 0.99 else 1,
                                  max_samples=params['max_samples'] if params['max_samples'] < 0.99 else 1,
                                  )

    model.fit(data['t_x'], data['t_y'].ravel())
    pred = model.predict(data['v_x'])
    mse = mean_squared_error(data['v_y'], pred, squared=False)
    return {'loss': mse, 'status': STATUS_OK, 'model': model}


def bayesian_tuning(space, model_type, max_evals=50, **data):
    score_func = None

    if model_type == 'xgbr':
        score_func = xgbr_score
    elif model_type == 'rfr':
        score_func = rfr_score

    trials = Trials()
    best = fmin(fn=partial(score_func, **data),
                space=space,
                algo=tpe.suggest,
                max_evals=max_evals,
                trials=trials)

    return best, trials



import pandas as pd


def get_submission_file(y_test, path):
    submission = pd.DataFrame(y_test, columns=['item_cnt_month']).reset_index()
    submission.to_csv(path, header=['ID', 'item_cnt_month'], index=False)
    return submission



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np


def target_prediction_scatter(target, pred):
    error_data = pd.DataFrame({'target': target, 'pred': pred})
    error_data['error'] = np.abs(target - pred)
    plt.figure()
    sns.scatterplot(data=error_data, x='pred', y='target', hue='error')
    plt.xlabel('Prediction')
    plt.ylabel('Target')
    plt.title('Target vs. Prediction dependency')
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    return plt.gcf()


def get_biggest_errors(val_x, val_y, val_pred, frac=0.01):
    val_with_errors = val_x.copy()
    val_with_errors['target'] = val_y
    val_with_errors['pred'] = val_pred
    val_with_errors['error'] = val_with_errors['target'] - val_with_errors['pred']
    val_with_errors['error^2'] = val_with_errors['error']**2
    val_with_errors.sort_values(by=['error^2'], inplace=True, ascending=False)
    return val_with_errors[:int(val_with_errors.shape[0]*frac)]


def get_problematic_categories(val_with_errors, val_orig, cols_list, k=5):
    val_orig = val_orig.loc[val_with_errors.index, :]
    problematic = {}
    for column in cols_list:
        problematic[column] = list(val_orig[column].value_counts().index[:k])

    return problematic


def plot_errors_across_target(val_with_errors, bins=[0, 1, 2, 10, 50, 100]):
    val_with_errors['target_binned'] = pd.cut(val_with_errors['target'], bins=bins)
    plt.figure()
    plt.title('Errors distribution across targets')
    sns.set(rc={'figure.figsize': (16, 11)})
    sns.violinplot(data=val_with_errors, x='target_binned', y='error')
    plt.axhline(y=0.0, color='r')
    return plt.gcf()




from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
import numpy as np
import pandas as pd
from datetime import date
from tqdm import tqdm


class MonthFromDate(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X['month'] = X['date_block_num'] % 12 + 1
        X['month_sin'] = np.sin(2 * np.pi * X['month'] / 12.0)
        X['month_cos'] = np.cos(2 * np.pi * X['month'] / 12.0)
        X.drop(['month'], axis=1, inplace=True)
        print('Month encoding stage complete')
        return X


class ShopItemAge(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X['shop_age_month'] = (X['date'] - X['first_shop_appearance']).dt.days / 30
        X['item_age_month'] = (X['date'] - X['first_item_appearance']).dt.days / 30
        X.drop(['first_shop_appearance', 'first_item_appearance', 'date'], axis=1, inplace=True)

        assert all(X['shop_age_month'] >= 0)
        assert all(X['item_age_month'] >= 0)
        assert not X.isna().any().any()

        print('Item and shop ages calculation complete')
        return X


class MonthlySales(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.groupby(['date_block_num', 'item_id', 'shop_id']).agg({'item_cnt_day': 'sum',
                                                                     'date': 'max',
                                                                     }).reset_index().rename(
            {'item_cnt_day': 'item_cnt_month'}, axis=1)
        print('Monthly sales calculated')
        return X


class MonthlySalesTest(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X['date_block_num'] = 34
        X['date'] = date(2015, 34 % 12 + 1, 28)
        X['date'] = pd.to_datetime(X['date'], format="%Y-%m-%d")
        print('Monthly sales calculated')
        return X


class MergeTables(BaseEstimator, TransformerMixin):
    def __init__(self, **kwargs):
        if not all(df in kwargs.keys() for df in ['items', 'shops', 'item_categories']):
            raise ValueError('Not enough dataframes provided')
        self._items = kwargs['items']
        self._shops = kwargs['shops']
        self._item_cat = kwargs['item_categories']

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        old_shape = X.shape[0]
        X = X.merge(self._items, on='item_id')  # , how='left'
        assert X.shape[0] == old_shape
        X = X.merge(self._shops, on='shop_id')  # , how='left'
        assert X.shape[0] == old_shape
        X = X.merge(self._item_cat, on='item_category_id')  # , how='left'
        assert X.shape[0] == old_shape
        assert not X.isna().any().any()

        X.drop([col for col in X.columns if X[col].dtype == object], axis=1, inplace=True)
        print('Merging stage complete')
        return X


# Lag feature transformers


class GetItemSalesLag(BaseEstimator, TransformerMixin):
    def __init__(self, lags, lag_data):
        self._lags = lags
        self._lag_data = lag_data

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        old_shape = X.shape[0]
        X = X.merge(self._lag_data, on=['item_id', 'shop_id'], how='left')  #
        assert old_shape == X.shape[0]
        X = X.fillna(0)

        for lag in self._lags:
            if date_block - lag >= 0:
                X[f'item_sales_lag_{lag}'] = X.loc[:, str(int(date_block - lag))]
            else:
                X[f'item_sales_lag_{lag}'] = 0

        return X


class GetPriceLag(BaseEstimator, TransformerMixin):
    def __init__(self, lags, lag_data):
        self._lags = lags
        self._lag_data = lag_data

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        old_shape = X.shape[0]
        X = X.merge(self._lag_data, on=['item_id', 'shop_id'], how='left')  #
        assert old_shape == X.shape[0]
        X = X.fillna(0)
        for lag in self._lags:
            if date_block - lag >= 0:
                X[f'price_lag_{lag}'] = X.loc[:, str(int(date_block - lag))]
            else:
                X[f'price_lag_{lag}'] = 0

        X.drop([str(i) for i in range(34)], axis=1, inplace=True)
        return X


class GetTotalSalesLag(BaseEstimator, TransformerMixin):
    def __init__(self, lags, lag_data):
        self._lags = lags
        self._lag_data = lag_data

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        old_shape = X.shape[0]
        X = X.merge(self._lag_data, on=['shop_id'], how='left')  #
        assert X.shape[0] == old_shape
        X = X.fillna(0)
        for lag in self._lags:
            if date_block - lag >= 0:
                X[f'total_sales_lag_{lag}'] = X.loc[:, str(int(date_block - lag))]
            else:
                X[f'total_sales_lag_{lag}'] = 0

        X.drop([str(i) for i in range(34)], axis=1, inplace=True)
        return X


class GetShopVarietyLag(BaseEstimator, TransformerMixin):
    def __init__(self, lags, lag_data):
        self._lags = lags
        self._lag_data = lag_data

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        old_shape = X.shape[0]
        X = X.merge(self._lag_data, on=['shop_id'], how='left')
        assert X.shape[0] == old_shape
        X = X.fillna(0)
        for lag in self._lags:
            if date_block - lag >= 0:
                X[f'shop_variety_lag_{lag}'] = X.loc[:, str(int(date_block - lag))]
            else:
                X[f'shop_variety_lag_{lag}'] = 0

        X.drop([str(i) for i in range(34)], axis=1, inplace=True)
        return X


class GetItemSpreadLag(BaseEstimator, TransformerMixin):
    def __init__(self, lags, lag_data):
        self._lags = lags
        self._lag_data = lag_data

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        old_shape = X.shape[0]
        X = X.merge(self._lag_data, on=['item_id'], how='left')  #
        assert X.shape[0] == old_shape
        X = X.fillna(0)
        for lag in self._lags:
            if date_block - lag >= 0:
                X[f'item_spread_lag_{lag}'] = X.loc[:, str(int(date_block - lag))]
            else:
                X[f'item_spread_lag_{lag}'] = 0

        X.drop([str(i) for i in range(34)], axis=1, inplace=True)
        return X


class GetSumOfSales(BaseEstimator, TransformerMixin):
    def __init__(self, lags):
        self._lags = lags

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        for lag in self._lags:
            cols_to_sum = [str(int(date_block - i)) for i in range(1, lag + 1) if date_block - i >= 0]
            X[f'sum_sales_lag_{lag}'] = X.loc[:, cols_to_sum].sum(axis=1)

        return X


class GetNonZeroMonths(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, y=None ):
        return self

    def transform(self, X, y=None):
        date_block = X['date_block_num'].values[0]
        prev_months = X[[str(i) for i in range(0, date_block)]]
        X['never_sold'] = np.count_nonzero(prev_months, axis=1) == 0
        X['never_sold'].replace({True: 1, False: 0}, inplace=True)
        X['mean_sales_when_sold'] = prev_months.apply(lambda row: row[row > 0].mean(), axis=1)
        X['mean_sales_when_sold'].fillna(0, inplace=True)
        X.drop(['date_block_num'], axis=1, inplace=True)
        X.drop([str(i) for i in range(34)], axis=1, inplace=True)

        return X


class FeatureExtractor:
    def __init__(self, feature_datasets: dict, lag_datasets: dict):
        self._feature_datasets = feature_datasets
        self._lag_datasets = lag_datasets

        self._extraction_pipeline_train = self.init_extraction_pipeline(test_set=False)
        self._extraction_pipeline_test = self.init_extraction_pipeline(test_set=True)
        self._lag_extraction_pipeline = self.init_lag_features_pipeline()

    def init_extraction_pipeline(self, test_set: bool = False) -> Pipeline:
        extraction_steps = [
            ("monthly_sales", MonthlySales()) if not test_set else ("monthly_sales", MonthlySalesTest()),
            ("month_feat", MonthFromDate()),
            ("merge", MergeTables(**self._feature_datasets)),
            ("ages", ShopItemAge())]

        return Pipeline(steps=extraction_steps)

    def init_lag_features_pipeline(self) -> Pipeline:
        lag_extraction_steps = [
            ("shop_variety", GetShopVarietyLag(lags=[1, 2, 3], lag_data=self._lag_datasets['shop_variety'])),
            ("item_spread", GetItemSpreadLag(lags=[1, 2, 3], lag_data=self._lag_datasets['item_spread'])),
            ("past_month_price", GetPriceLag(lags=[1, 2, 3], lag_data=self._lag_datasets['item_price'])),
            ("past_month_total_sales", GetTotalSalesLag(lags=[1, 3, 6], lag_data=self._lag_datasets['total_sales'])),
            (
                "past_month_item_sales",
                GetItemSalesLag(lags=[1, 2, 3, 6, 12], lag_data=self._lag_datasets['item_sales'])),
            ("sum_of_sales", GetSumOfSales(lags=[6, 12])),
            ("non-zero months", GetNonZeroMonths())
        ]

        return Pipeline(steps=lag_extraction_steps)

    def extract_features(self, train_data: pd.DataFrame=None, val_data: pd.DataFrame=None,
                               test_data: pd.DataFrame=None) -> (pd.DataFrame, pd.DataFrame):
        output = []
        if train_data is not None:
            train = self._extraction_pipeline_train.fit_transform(train_data)
            train_per_month = [
                self._lag_extraction_pipeline.fit_transform(train.loc[train['date_block_num'] == block].copy())
                for block in tqdm(train['date_block_num'].unique(), desc='Lag features extraction')]
            train = pd.concat(train_per_month, axis=0)
            output.append(train)
        if val_data is not None:
            val = self._extraction_pipeline_train.transform(val_data)
            val = self._lag_extraction_pipeline.transform(val)
            output.append(val)
        if test_data is not None:
            test = self._extraction_pipeline_test.fit_transform(test_data)
            test = self._lag_extraction_pipeline.transform(test)
            output.append(test)

        return output




import pandas as pd
import numpy as np
import os
import re
from datetime import date


def etl(src_path: os.PathLike, dst_path: os.PathLike = None, save: bool = False) -> None:
    sales = pd.read_csv(os.path.join(src_path, 'sales_train.csv'), parse_dates=['date'], dayfirst=True)
    shops = pd.read_csv(os.path.join(src_path, 'shops.csv'))
    items = pd.read_csv(os.path.join(src_path, 'items.csv'))
    item_cat = pd.read_csv(os.path.join(src_path, 'item_categories.csv'))
    test = pd.read_csv(os.path.join(src_path, 'test.csv'))
    sales = sales[sales['item_price'] > 0]
    sales = sales[sales['item_cnt_day'] > 0]

    items['item_name'] = items['item_name'].apply(
        lambda name: re.sub('^[\\\/^.*\[\]~!@#$%^&()_+={}|\:;“’<,>?฿]+', '', name))
    shops['shop_name'] = shops['shop_name'].apply(
        lambda name: re.sub('^[\\\/^.*\[\]~!@#$%^&()_+={}|\:;“’<,>?฿]+', '', name))

    shops_mapping = {10: 11, 40: 39, 0: 57, 1: 58}
    items_mapping = {12: 14690}

    sales['shop_id'].replace(shops_mapping, inplace=True)
    sales['item_id'].replace(items_mapping, inplace=True)

    test['item_id'].replace(items_mapping, inplace=True)
    test['shop_id'].replace(shops_mapping, inplace=True)

    sales = sales.groupby(['date', 'date_block_num', 'shop_id', 'item_id']).agg(
        {'item_price': 'median', 'item_cnt_day': 'sum'}).reset_index()

    if save:
        os.mkdir(dst_path)
        sales.to_csv(os.path.join(dst_path, 'sales_train.csv'), index=False, date_format='%d.%m.%Y')
        shops.to_csv(os.path.join(dst_path, 'shops.csv'), index=False)
        items.to_csv(os.path.join(dst_path, 'items.csv'), index=False)
        item_cat.to_csv(os.path.join(dst_path, 'item_categories.csv'), index=False)
        test.to_csv(os.path.join(dst_path, 'test.csv'), index=False)
    else:
        return {'sales': sales,
                'shops': shops,
                'items': items,
                'item_cat': item_cat,
                'test': test}


def resample_sales(sales, shops, items,
                   zeros_frac=0.1, big_targets_frac=0.1, big_target_thres=0.95, drop_ones_frac=0.5):
    def get_random_dates(n):
        years = np.random.choice([2013, 2015], size=n, replace=True)
        months = np.random.choice(np.arange(1, 13), size=n, replace=True)
        days = np.random.choice(np.arange(1, 29), size=n, replace=True)
        dates = list(map(lambda year, month, day: date(year, month, day), years, months, days))
        return dates

    shop_item_merged = shops.merge(items, how='cross')
    possible_shop_item_pairs = all_shop_item_pairs = set(zip(shop_item_merged['shop_id'], shop_item_merged['item_id']))
    real_shop_item_pairs = set(zip(sales['shop_id'], sales['item_id']))
    artificial_zero_sales_pairs = np.array([*(possible_shop_item_pairs - real_shop_item_pairs)])

    price_per_category = sales.merge(items, on='item_id').groupby(['item_category_id']).agg(
        {'item_price': 'median'}).reset_index()

    zero_sales_n = int(sales.shape[0] * zeros_frac)
    zero_sales = pd.DataFrame()
    zero_sales['date'] = pd.to_datetime(get_random_dates(zero_sales_n), format="%Y-%m-%d")
    zero_sales['date_block_num'] = (zero_sales['date'].dt.year - 2013) * 12 + zero_sales[
        'date'].dt.month + 1
    zero_sales_pairs = artificial_zero_sales_pairs[
                       np.random.randint(artificial_zero_sales_pairs.shape[0], size=zero_sales_n), :]
    zero_sales['shop_id'] = zero_sales_pairs[:, 0]
    zero_sales['item_id'] = zero_sales_pairs[:, 1]
    zero_sales = zero_sales.merge(items, on='item_id').merge(price_per_category, on='item_category_id')
    zero_sales['item_cnt_day'] = 0
    zero_sales = zero_sales[sales.columns]
    zero_sales = zero_sales[zero_sales['date_block_num'] < 34]

    big_targets = sales[sales['item_cnt_day'] > sales['item_cnt_day'].quantile(big_target_thres)]
    big_target_shops = big_targets['shop_id'].unique()
    big_target_items = big_targets['item_id'].unique()
    big_target_targets = big_targets['item_cnt_day'].unique()

    big_sales_n = int(sales.shape[0] * big_targets_frac)
    big_sales = pd.DataFrame()
    big_sales['date'] = pd.to_datetime(get_random_dates(big_sales_n), format="%Y-%m-%d")
    big_sales['date_block_num'] = (big_sales['date'].dt.year - 2013) * 12 + big_sales['date'].dt.month + 1
    big_sales['shop_id'] = np.random.choice(big_target_shops, size=big_sales_n, replace=True)
    big_sales['item_id'] = np.random.choice(big_target_items, size=big_sales_n, replace=True)
    big_sales = big_sales.merge(items, on='item_id').merge(price_per_category, on='item_category_id')
    big_sales['item_cnt_day'] = np.random.choice(big_target_targets, size=big_sales_n, replace=True)
    big_sales = big_sales[sales.columns]

    new_sales = pd.concat([sales, zero_sales, big_sales])
    new_sales = new_sales.drop(new_sales[new_sales['item_cnt_day'] == 1.0].sample(frac=drop_ones_frac).index)
    new_sales = new_sales[new_sales['date_block_num'] < 34]

    return new_sales



import pandas as pd
import re
import os


class FeatureGenerator:
    def __init__(self, dest_dir: str = None, save: bool = False, **datasets):
        if save:
            self._dst = dest_dir
            os.mkdir(self._dst)
        else:
            self._dst = None

        if not all(df in datasets.keys() for df in ['items', 'shops', 'item_categories', 'sales']):
            raise ValueError('Not enough dataframes provided')

        self._shops = datasets['shops']
        self._items = datasets['items']
        self._sales = datasets['sales']
        self._item_cat = datasets['item_categories']

    def saturate_shops(self) -> None:
        self._shops['city_name'] = self._shops['shop_name'].apply(
            lambda s: re.split(' (ТРЦ|ТЦ|ТК|ТРК|МТРЦ|\(|\"|ул)', s)[0])
        self._shops['city_name'].replace({'Якутск Орджоникидзе, 56': 'Якутск', 'Москва Магазин С21': 'Москва'},
                                         inplace=True)
        self._shops['city_id'] = self._shops['city_name'].rank(method='min').astype(int)

        shop_appearance = self._sales.groupby('shop_id').agg({'date': 'min'}).reset_index(). \
            rename(columns={'date': 'first_shop_appearance'})
        shops_saturated = self._shops.merge(shop_appearance, on='shop_id', how='left')
        shops_saturated.fillna(self._sales['date'].max(), inplace=True)

        if self._dst is not None:
            shops_saturated.to_csv(os.path.join(self._dst, 'shops.csv'), index=False, date_format='%d.%m.%Y')
        else:
            return shops_saturated

    def saturate_items(self) -> None:
        item_appearance = self._sales.groupby('item_id').agg({'date': 'min'}).reset_index(). \
            rename(columns={'date': 'first_item_appearance'})
        items_saturated = self._items.merge(item_appearance, on='item_id', how='left')
        items_saturated.fillna(self._sales['date'].max(), inplace=True)
        if self._dst is not None:
            items_saturated.to_csv(os.path.join(self._dst, 'items.csv'), index=False, date_format='%d.%m.%Y')
        else:
            return items_saturated

    def saturate_categories(self) -> None:
        self._item_cat['item_global_category_name'] = self._item_cat['item_category_name']. \
            apply(lambda s: re.split(' (-|\()', s)[0])
        self._item_cat['item_global_category_id'] = self._item_cat['item_global_category_name'].rank(
            method='min').astype(int)

        if self._dst is not None:
            self._item_cat.to_csv(os.path.join(self._dst, 'item_categories.csv'), index=False)
        else:
            return self._item_cat

    def get_item_sales_lags(self) -> None:
        sales_monthly = self._sales.groupby(['date_block_num', 'item_id', 'shop_id']).agg({'item_cnt_day': 'sum'}). \
            reset_index().rename({'item_cnt_day': 'item_cnt_month'}, axis=1)
        sales_monthly_pivot = pd.pivot_table(sales_monthly,
                                             values='item_cnt_month',
                                             index=['item_id', 'shop_id'],
                                             columns=['date_block_num']).fillna(0).reset_index()
        if self._dst is not None:
            sales_monthly_pivot.to_csv(os.path.join(self._dst, 'item_sales_monthly.csv'), index=False)
        else:
            return sales_monthly

    def get_price_lags(self) -> None:
        price_monthly = self._sales.groupby(['date_block_num', 'item_id', 'shop_id']).agg(
            {'item_price': 'median'}).reset_index()
        price_monthly_pivot = pd.pivot_table(price_monthly, values='item_price', index=['item_id', 'shop_id'],
                                             columns=['date_block_num']).fillna(0).reset_index()
        if self._dst is not None:
            price_monthly_pivot.to_csv(os.path.join(self._dst, 'item_price_monthly.csv'), index=False)
        else:
            return price_monthly_pivot

    def get_total_sales_lags(self) -> None:
        total_sales_monthly = self._sales.groupby(['date_block_num', 'shop_id']).agg(
            {'item_cnt_day': 'sum'}).reset_index().rename(
            columns={'item_cnt_day': 'item_sold_month'})
        total_sales_monthly_pivot = pd.pivot_table(total_sales_monthly, values='item_sold_month', index=['shop_id'],
                                                   columns=['date_block_num']).fillna(0).reset_index()
        if self._dst is not None:
            total_sales_monthly_pivot.to_csv(os.path.join(self._dst, 'total_sales_monthly.csv'), index=False)
        else:
            return total_sales_monthly_pivot

    def get_shop_variety(self) -> None:  # unique items sold in a shop per month
        shop_variety = self._sales.groupby(['shop_id', 'date_block_num']).agg({'item_id': 'nunique'}).reset_index()
        shop_variety.rename(columns={'item_id': 'shop_variety'}, inplace=True)
        shop_variety_pivot = pd.pivot_table(shop_variety, values='shop_variety', index=['shop_id'],
                                            columns=['date_block_num']).fillna(0).reset_index()
        if self._dst is not None:
            shop_variety_pivot.to_csv(os.path.join(self._dst, 'shop_variety_monthly.csv'), index=False)
        else:
            return shop_variety_pivot

    def get_item_spread(self) -> None:  # unique shops that sell an item per month
        item_spread = self._sales.groupby(['item_id', 'date_block_num']).agg({'shop_id': 'nunique'}).reset_index()
        item_spread.rename(columns={'shop_id': 'item_spread'}, inplace=True)
        item_spread_pivot = pd.pivot_table(item_spread, values='item_spread', index=['item_id'],
                                           columns=['date_block_num']).fillna(0).reset_index()
        if self._dst is not None:
            item_spread_pivot.to_csv(os.path.join(self._dst, 'item_spread_monthly.csv'), index=False)
        else:
            return item_spread_pivot



import shap
import matplotlib.pyplot as plt


class ShapExplainer:
    def __init__(self, model):
        shap.initjs()
        self._explainer = shap.TreeExplainer(model)
        self._shap_values = None
        self._data = None

    def get_shap_values(self, data):
        self._data = data
        self._shap_values = self._explainer.shap_values(data)
        return self._shap_values

    def summary_plot(self, max_display=20):
        plt.figure()
        shap.summary_plot(self._shap_values, self._data, feature_names=self._data.columns, max_display=max_display,
                          show=False)
        return plt.gcf()

    def plot_prediction(self, data, feature_names):
        plt.figure()
        error_shap_value = self._explainer.shap_values(data.values.reshape(1, -1))
        plot = shap.force_plot(self._explainer.expected_value, error_shap_value, data.values,
                               feature_names=feature_names, show=False,
                               matplotlib=True)
        return plot



import pandas as pd


def run_quality_check(**kwargs: pd.DataFrame) -> None:
    pd.set_option('display.float_format', lambda x: '%.5f' % x)

    if not all(df in kwargs.keys() for df in ['sales', 'items', 'shops', 'item_categories', 'test']):
        raise ValueError('Not enough dataframes provided')
    else:
        for name, df in kwargs.items():
            if df.isna().any().any():
                print(f'NA values in [{name}]')
            if df.duplicated().any():
                print(f'Duplicates in [{name}]')
                print(df[df.duplicated()])
        print(kwargs['sales']['date'].head())
        # sales quality
        if not pd.api.types.is_datetime64_dtype(kwargs['sales']['date'].dtype):
            print('[sales] Wrong date dtype')
        else:
            if (kwargs['sales']['date'].min().year != 2013 and kwargs['sales']['date'].min().month != 1 and
                    kwargs['sales']['date'].max().year != 2015 and kwargs['sales']['date'].max().month != 10):
                print('[sales] Incorrect date range')

        if kwargs['sales']['date_block_num'].nunique() != 34:
            print('[sales] Date blocks missing')

        print('[sales] item price statistics:')
        print(kwargs['sales']['item_price'].describe().loc[['min', 'max', 'mean', '50%']])
        print('[sales] item_cnt_day statistics:')
        print(kwargs['sales']['item_cnt_day'].describe().loc[['min', 'max', 'mean', '50%']])

        # items quality
        bad_name_items = kwargs['items']['item_name'].str.contains('^[\\\/^.*\[\]~!@#$%^&()_+={}|\:;“’<,>?฿]+.*',
                                                                  regex=True, na=False).sum()
        if bad_name_items > 0:
            print(f'[items] Item names contain special characters ({bad_name_items} cases)')

        # shops quality
        bad_name_shops = kwargs['shops']['shop_name'].str.contains('^[\\\/^.*\[\]~!@#$%^&()_+={}|\:;“’<,>?฿]+.*',
                                                                   regex=True, na=False).sum()
        if bad_name_shops > 0:
            print(f'[shops] Shop names contain special characters ({bad_name_shops} cases)')

        # item categories quality
        bad_name_categories = kwargs['item_categories']['item_category_name'].str.contains('^[\\\/^.*\[\]~!@#$%^&()_+={}|\:;“’<,>?฿]+.*',
                                                                   regex=True, na=False).sum()
        if bad_name_categories > 0:
            print(f'[item_categories] Category names contain special characters ({bad_name_shops} cases)')

        # joint keys check
        shop_ids_sales = set(kwargs['sales']['shop_id'])
        item_ids_sales = set(kwargs['sales']['item_id'])
        shop_ids_shops = set(kwargs['shops']['shop_id'])
        item_ids_items = set(kwargs['items']['item_id'])
        shop_ids_test = set(kwargs['test']['shop_id'])
        item_ids_test = set(kwargs['test']['item_id'])
        cat_ids_items = set(kwargs['items']['item_category_id'])
        cat_ids_categories = set(kwargs['item_categories']['item_category_id'])
        if len(shop_ids_sales - shop_ids_shops) > 0:
            print('shops in [sales] not present in [shops]')
        if len(item_ids_sales - item_ids_items) > 0:
            print('items in [sales] not present in [items]')
        if len(shop_ids_test - shop_ids_shops) > 0:
            print('shops in [test] not present in [shops]')
        if len(item_ids_test - item_ids_items) > 0:
            print('items in [test] not present in [items]')
        if len(cat_ids_items - cat_ids_categories) > 0:
            print('categories in [items] not present in [item_categories]')











import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline

from sklearn.preprocessing import StandardScaler
import category_encoders as ce


class FeatureScaler(BaseEstimator, TransformerMixin):
    def __init__(self):
        self._feat_scaler = StandardScaler()
        self._num_cols = None

    def fit(self, X, y=None):
        self._num_cols = list(set(X.select_dtypes(include='float64').columns) - {'month_sin', 'month_cos'})
        self._feat_scaler = self._feat_scaler.fit(X.loc[:, self._num_cols])
        return self

    def transform(self, X, y=None):
        X.loc[:, self._num_cols] = self._feat_scaler.transform(X.loc[:, self._num_cols])
        return X


class TargetEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, smoothing=1.0, min_samples_leaf=1):
        self._encoder = ce.TargetEncoder(min_samples_leaf=min_samples_leaf, smoothing=smoothing)
        self._cat_cols = ['shop_id', 'city_id', 'item_id', 'item_category_id', 'item_global_category_id']

    def fit(self, X, y=None):
        X.loc[:, self._cat_cols] = X.loc[:, self._cat_cols].astype(object)
        self._encoder = self._encoder.fit(X.loc[:, self._cat_cols], y)
        return self

    def transform(self, X, y=None):
        X.loc[:, self._cat_cols] = self._encoder.transform(X.loc[:, self._cat_cols])
        return X


class TargetTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, no_log=False):
        self._scaler = StandardScaler()
        self._no_log = no_log

    def fit(self, X, y=None):
        if self._no_log:
            self._scaler = self._scaler.fit(X)
        else:
            self._scaler = self._scaler.fit(np.log(X))
        return self

    def transform(self, X, y=None):
        if not self._no_log:
            X = np.log(X)
        X = self._scaler.transform(X)
        return X

    def inverse_transform(self, X, y=None):
        X = self._scaler.inverse_transform(X)
        if not self._no_log:
            X = np.exp(X)
        return X


class OutlierRemover(BaseEstimator, TransformerMixin):
    def __init__(self, target='item_cnt_month', prob=0.03):
        self._prob = prob
        self._threshold = None
        self._target = target

    def fit(self, X, y=None):
        self._threshold = np.mean(y) / self._prob
        return self

    def fit_transform(self, X, y=None):
        self._threshold = np.mean(y) / self._prob
        X = X[y < self._threshold]
        y = y[y < self._threshold]
        return X, y

    def transform(self, X, y=None):
        X = X[y < self._threshold]
        y = y[y < self._threshold]
        return X, y


class Preprocessor:
    def __init__(self, target_col, no_log_target=False, outlier_prob=0.03, keep_cat=False, **encoder_params):
        self.target_transformer = TargetTransformer(no_log=no_log_target)
        self._outlier_remover = OutlierRemover(prob=outlier_prob, target=target_col)
        self._target_col = target_col

        preprocessing_steps = [('scaler', FeatureScaler())]
        if not keep_cat:
            preprocessing_steps.append(('encoder', TargetEncoder(**encoder_params)))

        self._feature_prep_pipeline = Pipeline(steps=preprocessing_steps)

    def preprocess_data(self, train_data=None, val_data=None, test_data=None):

        output = []
        if train_data is not None:
            train_x, train_y = train_data
            train_x, train_y = self._outlier_remover.fit_transform(train_x, train_y)
            train_y_transformed = self.target_transformer.fit_transform(train_y.values.reshape(-1, 1))
            train_x = self._feature_prep_pipeline.fit_transform(train_x, train_y)
            output.append(train_x)
            output.append(pd.DataFrame(train_y_transformed))
            output.append(pd.DataFrame(train_y))
        if val_data is not None:
            val_x, val_y = val_data
            val_x, val_y = self._outlier_remover.transform(val_x, val_y)
            val_y_transformed = self.target_transformer.transform(val_y.values.reshape(-1, 1))
            assert (val_x.columns == train_x.columns).all()
            assert type(val_x) == type(train_x)
            val_x = self._feature_prep_pipeline.transform(val_x)
            output.append(val_x)
            output.append(pd.DataFrame(val_y_transformed))
            output.append(pd.DataFrame(val_y))
        if test_data is not None:
            test_x = self._feature_prep_pipeline.transform(test_data)
            output.append(test_x)

        return output



from .data import DQC, ETL
from .features import feature_engineering, feature_extractor, preprocessing, validation_schema
from .models import tuning, train_model, feature_selection, submission
from .analysis import error_analysis, shap_explainer
from .version import __version__




import pandas as pd
from datetime import date, timedelta


class ValidationSchema:
    def __init__(self, data):
        if 'date' not in data.columns:
            raise ValueError('Dataset doesnt have ''date'' feature')
        self._data = data
        self._train_start = data['date'].min()
        self._val_start = data['date'].max() - timedelta(days=30)  # last month
        self._val_end = data['date'].max()

    def split_data(self, train_start: date = None, val_start: date = None) -> (pd.DataFrame, pd.DataFrame):
        if train_start is not None:
            self._train_start = pd.Timestamp(train_start)
        if val_start is not None:
            self._val_start = pd.Timestamp(val_start)
            self._val_end = pd.Timestamp(val_start + timedelta(days=30))

        if train_start > val_start:
            raise ValueError('Validation fold can''t be located before training fold')
        if (val_start - train_start).days < 365:
            raise ValueError('Train start and validation start must be at least a year apart')

        return self._data[(self._data['date'] >= self._train_start) & (self._data['date'] < self._val_start)], \
            self._data[(self._data['date'] >= self._val_start) & (self._data['date'] < self._val_end)]

    def get_val(self, val_start: date) -> pd.DataFrame:
        val_start = pd.Timestamp(val_start)
        val_end = val_start + timedelta(days=30)

        return self._data[(self._data['date'] >= val_start) & (self._data['date'] < val_end)]

    @staticmethod
    def x_y_split(train: pd.DataFrame, val: pd.DataFrame, target: str) -> (pd.DataFrame, pd.DataFrame,
                                                                                pd.DataFrame, pd.DataFrame):
        return train.drop(columns=[target], axis=1), train[target], \
               val.drop(columns=[target], axis=1), val[target]



import pandas as pd
import numpy as np
import os
import sys

from sklearn.metrics import mean_squared_error


def fit_eval_model(model, t_x, t_y, v_x, v_y, target_transformer):
    model.fit(t_x, t_y.ravel())
    train_pred = target_transformer.inverse_transform(model.predict(t_x))
    train_rmse = mean_squared_error(train_y, train_pred, squared=False)
    val_pred = target_transformer.inverse_transform(model.predict(v_x))
    val_rmse = mean_squared_error(val_y, val_pred, squared=False)
    print(f'train RMSE - {np.round(train_rmse, 3)}, val RMSE - {np.round(val_rmse,3)}')
    return model, train_rmse, val_rmse


if __name__ == '__main__':
    train_x = pd.read_csv(os.path.join(sys.argv[1], 'train_x.csv'))
    val_x = pd.read_csv(os.path.join(sys.argv[1], 'val_x.csv'))
    train_y = pd.read_csv(os.path.join(sys.argv[1], 'train_y.csv'))
    val_y = pd.read_csv(os.path.join(sys.argv[1], 'val_y.csv'))




